var vows = require('vows')
  , assert = require('assert')
  , Emitter = require("events").EventEmitter
  , Ax = require("ax")
  , log = new Ax({ level: "debug", file: "log/specs/shred.log" })
  , Shred = require("../lib/shred")
;

vows.describe('Shred').addBatch({
  'A minimal valid GET request': {
    topic: function() {
      
      var shred = new Shred({ logger: log })
        , promise = new(Emitter)
      ;
      
      var req = shred.get({
        url: "http://localhost:1337/200",
        on: {
          response: function(response) {
            promise.emit("success", response);
          },
          error: function(error) {
            log.debug(error);
            log.info("Is rephraser running?")
          }
        }
      });
      return promise;
    },
    "should have a response status code of 200": function(response){
      assert.equal(response.status, 200);
    },
    "should have no content type": function(response) {
      assert.equal(!!response.content.data.headers["Content-Type"],false);
    },
  },
  'A minimal valid POST request': {
    topic: function() {
      
      var shred = new Shred({ logger: log })
        , promise = new(Emitter)
      ;
      
      shred.post({
        url: "http://localhost:1337/200",
        on: {
          response: function(response) {
            promise.emit("success",response);
          },
          error: function(error) {
            log.debug(error);
            log.info("Is rephraser running?")
          }
        }
      });
      
      return promise;
    },
    "should have a response status code of 200": function(response){
      assert.equal(response.status, 200);
    },
    "should have no content type": function(response) {
      assert.equal(!!response.content.data.headers["Content-Type"],false);
    }
  },
  'A POST request with a body': {
    topic: function() {
      
      var shred = new Shred({ logger: log })
        , promise = new(Emitter)
      ;
      
      shred.post({
        url: "http://localhost:1337/200",
        body: "Hello",
        on: {
          response: function(response) {
            promise.emit("success",response);
          },
          error: function(error) {
            log.debug(error);
            log.info("Is rephraser running?")
          }
        }
      });
      
      return promise;
    },
    "should have a response status code of 200": function(response){
      assert.equal(response.status, 200);
    },
    "should have a default content type of 'text/plain'": function(response) {
      assert.equal(
        response.content.data.headers["Content-Type"],
        "text/plain");
    }
  },
  "A POST request with a content type of 'application/json body that returns a 201": {
    topic: function() {
      
      var shred = new Shred({ logger: log })
        , promise = new(Emitter)
      ;
      shred.post({
        url: "http://localhost:1337/201",
        body: {foo: 1, bar: 2},
        on: {
          response: function(response) {
            promise.emit("success",response);
          },
          error: function(error) {
            log.debug(error);
            log.info("Is rephraser running?")
          }
        }
      });
      
      return promise;
    },
    "should have a response status code of 201": function(response){
      assert.equal(response.status, 201);
    },
    "should include a Location header in the response": function(response) {
      assert.equal(!!response.getHeader("Location"),true);
    },
    "should have a default content type of 'application/json'":
      function(response) {
      assert.equal(
        response.content.data.headers["Content-Type"],
        "application/json");
    }
  },
  "A GET request to a redirected URL": {
    topic: function() {
      
      var shred = new Shred({ logger: log })
        , promise = new(Emitter)
      ;
      shred.get({
        url: "http://localhost:1337/301",
        on: {
          response: function(response) {
            promise.emit("success",response);
          },
          error: function(error) {
            log.debug(error);
            log.info("Is rephraser running?")
          }
        }
      });
      
      return promise;
    },
    "will transparently handle the redirect": function(response){
      assert.equal(response.status, 200);
    }
  },
  "A request with an event handler based on the status code": {
    topic: function() { 
      var shred = new Shred({ logger: log })
        , promise = new(Emitter)
      ;
      
      var handleCount = 0;
      
      var req = shred.get({
        url: "http://localhost:1337/200",
        on: {
          200: function(response) {
            handleCount++;
            if (handleCount == 2) {
              promise.emit("success", response, handleCount);
            }
          },
          error: function(error) {
            log.debug(error);
            log.info("Is rephraser running?")
          }
        }
      });
      
      req.on(200, function(response) {
        handleCount++;
        if (handleCount == 2) {
          promise.emit("success", response, handleCount);
        }
      })
      
      return promise;
    },
    "will trigger the correct callback": function(response) {
      assert.equal(response.status,200);
    },
    "can have multiple callbacks": function(err, response, count) {
      assert.equal(count, 2);
    }
  },
  'A GET request with multiple handlers': {
    topic: function() {
      
      var shred = new Shred({ logger: log })
        , promise = new(Emitter)
      ;
      
      var handleRunCount = 0;
      var numberOfHandlers = 0;

      var createHandlerFunction = function () {
        numberOfHandlers ++;
        return function (response) {
          handleRunCount++;
          if (handleRunCount == numberOfHandlers) {
            promise.emit("success", response, numberOfHandlers, handleRunCount);
          }
        };
      };

      var req = shred.get({
        url: "http://localhost:1337/200",
        // Handler defined in the request options hash
        on: {
          response: createHandlerFunction(),
          error: function(error) {
            log.debug(error);
            log.info("Is rephraser running?")
          }
        }
      });

      // Handler without an event name
      req.on(createHandlerFunction());

      // Handler with an event name
      req.on('response', createHandlerFunction());

      // Handler with a hash of event names
      req.on({
        response: createHandlerFunction()
      });

      return promise;
    },
    "should be able to have multiple handlers": function(err, response, numberOfHandlers, handleRunCount) {
      // Sanity check to make sure that there are handlers:
      assert.notEqual(0, numberOfHandlers);
      assert.notEqual(1, numberOfHandlers);
    },
    "should run all of the handlers": function (err, response, numberOfHandlers, handleRunCount) {
      assert.equal(handleRunCount, numberOfHandlers);
    }
  }
}).export(module);
